# GNU Solfege - eartraining for GNOME
# Copyright (C) 2000, 2001, 2002  Tom Cato Amundsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
>>> a=MusicalPitch()
>>> a.m_octave_i == a.m_accidental_i == a.m_octave_i == 0
1
>>> a.semitone_pitch()
48
>>> a.get_octave_notename()
'c'
>>> (a+2).get_octave_notename()
'd'
>>> (2+a).get_octave_notename()
'd'
>>> MusicalPitch.new_from_notename("des'").get_notename()
'des'
>>> MusicalPitch.new_from_int(50).semitone_pitch()
50
>>> MusicalPitch.new_from_int(50).get_notename()
'd'
"""
import types, random
import locale

if __name__ == '__main__':
    import gettext
    gettext.install('solfege', 'share/locale', unicode=1)
    _i = _

# The following are here so that the strings are catched by pygettext
_("notename|c")
_("notename|ces")
_("notename|ceses")
_("notename|cis")
_("notename|cisis")
_("notename|d")
_("notename|des")
_("notename|deses")
_("notename|dis")
_("notename|disis")
_("notename|e")
_("notename|ees")
_("notename|eeses")
_("notename|eis")
_("notename|eisis")
_("notename|f")
_("notename|fes")
_("notename|feses")
_("notename|fis")
_("notename|fisis")
_("notename|g")
_("notename|ges")
_("notename|geses")
_("notename|gis")
_("notename|gisis")
_("notename|a")
_("notename|aes")
_("notename|aeses")
_("notename|ais")
_("notename|aisis")
_("notename|b")
_("notename|bes")
_("notename|beses")
_("notename|bis")
_("notename|bisis")

class InvalidNotenameException:
    def __init__(self, n):
        self.m_notename = n
    def __str__(self):
        return self.m_notename

class MusicalPitch:
    LOWEST_STEPS = -28
    HIGHEST_STEPS = 47
    def clone(self):
        r = MusicalPitch()
        r.m_octave_i = self.m_octave_i
        r.m_notename_i = self.m_notename_i
        r.m_accidental_i = self.m_accidental_i
        return r
    def new_from_notename(n):
        assert type(n) == type("")
        r = MusicalPitch()
        r.set_from_notename(n)
        return r
    new_from_notename = staticmethod(new_from_notename)
    def new_from_int(i):
        assert type(i) == type(0)
        r = MusicalPitch()
        r.set_from_int(i)
        return r
    new_from_int = staticmethod(new_from_int)
    def __init__(self):
        """
         c,,,, is lowest: m_octave_i == -4, steps() == -28
         g'''''' is highest: m_octave_i = 6, steps() == 46
        """
        self.m_octave_i = self.m_accidental_i = self.m_notename_i = 0
    def transpose_by_musicalpitch(self, P):
        """Silly function used by mpd/parser.py and company
        (d') transposes up one major second.
        """
        tra = P.semitone_pitch() - 60
        old_p = self.semitone_pitch()
        self.m_notename_i = self.m_notename_i + P.m_notename_i
        self.m_accidental_i = self.m_accidental_i + P.m_accidental_i
        if self.m_notename_i > 6:
            self.m_notename_i = self.m_notename_i - 7
            self.m_octave_i = self.m_octave_i + 1
        self.m_octave_i = self.m_octave_i + P.m_octave_i - 1
        if self.semitone_pitch()-old_p < tra:
            self.m_accidental_i = self.m_accidental_i + 1
        elif self.semitone_pitch()-old_p > tra:
            self.m_accidental_i = self.m_accidental_i - 1
        return self
    def steps(self):
        return self.m_notename_i + self.m_octave_i * 7
    def semitone_pitch(self):
        return [0, 2, 4, 5, 7, 9, 11][self.m_notename_i] + \
               self.m_accidental_i + self.m_octave_i * 12 + 48
    def set_from_int(self, midiint):
        self.m_octave_i = (midiint-48)/12
        self.m_notename_i = {0:0, 1:0, 2:1, 3:1, 4:2, 5:3, 6:3, 7:4, 8:4,
                             9:5, 10:5, 11:6}[midiint % 12]
        self.m_accidental_i = midiint-(self.m_octave_i+4)*12 \
                              -[0, 2, 4, 5, 7, 9, 11][self.m_notename_i]
    def set_from_notename(self, notename):
        tmp = notename
        self.m_accidental_i = self.m_octave_i = 0
        while notename[-1] in ["'", ","]:
            if notename[-1] == "'":
                self.m_octave_i = self.m_octave_i + 1
            elif notename[-1] == ",":
                self.m_octave_i = self.m_octave_i - 1
            notename = notename[:-1]
        if notename[:2] == 'es':
            notename = 'ees' + notename[2:]
        if notename[:2] == 'as':
            notename = 'aes' + notename[2:]
        while notename[-2:] == 'es':
            self.m_accidental_i = self.m_accidental_i -1
            notename = notename[:-2]
        while notename[-2:] == 'is':
            self.m_accidental_i = self.m_accidental_i + 1
            notename = notename[:-2]
        try:
            self.m_notename_i = ['c', 'd', 'e', 'f', 'g', 'a', 'b'].index(notename)
        except:
           raise InvalidNotenameException(tmp)
    def randomize(self, lowest, highest):
        """
        lowest and highest can be an integer, string or a MusicalPitch instance
        """
        #FIXME should we only accept strings as argument to randomize??
        assert type(lowest) == type(highest)
        if type(lowest) == types.StringType:
            lowest = MusicalPitch.new_from_notename(lowest).semitone_pitch()
        if type(highest) == types.StringType:
            highest = MusicalPitch.new_from_notename(highest).semitone_pitch()
        self.set_from_int(random.randint(int(lowest), int(highest)))
        return self
    def __radd__(self, a):
        return self + a
    def __add__(self, i):
        """
        MusicalPitch + integer = MusicalPitch
        MusicalPitch + Interval = MusicalPitch
        """
        if type(i) == type(0):
            v = self.semitone_pitch()
            assert 0 <= v + i < 128
            return MusicalPitch.new_from_int(v+i)
        elif i.__class__.__name__ == 'Interval':#isinstance(i, interval.Interval):
            r = self.clone()
            _p = r.semitone_pitch()
            r.m_notename_i = r.m_notename_i + i.m_interval * i.m_dir
            r.m_octave_i = r.m_octave_i + r.m_notename_i / 7 + i.m_octave * i.m_dir
            r.m_notename_i = r.m_notename_i % 7
            _diff = r.semitone_pitch() - _p
            r.m_accidental_i = r.m_accidental_i + (i.get_intvalue() - _diff)
            # to avoid notenames like ciscisciscis :
            if r.m_accidental_i > 2:
               r.m_accidental_i = r.m_accidental_i - 2
               r.m_notename_i = r.m_notename_i + 1
               if r.m_notename_i == 7:
                   r.m_notename_i = 0
                   r.m_octave_i = r.m_octave_i + 1
            if r.m_accidental_i < -2:
               r.m_accidental_i = r.m_accidental_i + 2
               r.m_notename_i = r.m_notename_i - 1
               if r.m_notename_i == -1:
                   r.m_notename_i = 6
                   r.m_octave_i = r.m_octave_i - 1
            return r
        else:
            raise "mpd.MusicalPitch.__add__"
    def __sub__(self, i):
        """
        MusicalPitch - MusicalPitch = integer
        MusicalPitch - integer = MusicalPitch
        """
        if isinstance(i, MusicalPitch):
            return self.semitone_pitch() - i.semitone_pitch()
        assert type(i) == types.IntType
        v = self.semitone_pitch()
        assert 0 <= v - i < 128
        return MusicalPitch.new_from_int(v-i)
    def __int__(self):
        return self.semitone_pitch()
    def __cmp__(self, B):
        if (self is None or self is None):
            return -1
        diff = self - B
        if diff < 0:
            return -1
        elif diff > 0:
            return 1
        else:
            return 0
    def __str__(self):
        return "(MusicalPitch %s)" % self.get_octave_notename()
    def get_user_notename(self):
        return self._get_notename(1, 0)
    def get_user_octave_notename(self):
        return self._get_notename(1, 1)
    def get_notename(self):
        return self._get_notename(0, 0)
    def get_octave_notename(self):
        return self._get_notename(0, 1)
    def _get_notename(self, user, octave):
        """
        user=0 for solfege internal notenames
        user=1 for notenames suited for users
        octave BOOLEAN show octave
        """
        s = ['c', 'd', 'e', 'f', 'g', 'a', 'b'][self.m_notename_i]
        assert -3 < self.m_accidental_i < 3, self.m_accidental_i
        s += ['eses', 'es', '', 'is', 'isis'][self.m_accidental_i+2]
        if user:
            s = "notename|" + s
            s = _i(s)
        if octave:
            if user and locale.getdefaultlocale()[0] == 'no_NO':
                if self.m_octave_i > 0:
                    s += "<sup>%s</sup>" % self.m_octave_i
                elif self.m_octave_i < 0:
                    s = "<sub>%s</sub>" % (-self.m_octave_i) + s.capitalize()
            else:
                if self.m_octave_i < 0:
                    s = s + ',' * (-self.m_octave_i)
                if self.m_octave_i > 0:
                    s = s + "'" * self.m_octave_i
        return s

def _test():
    import doctest
    import musicalpitch
    return doctest.testmod(musicalpitch)

if __name__ == "__main__":
    _test()
    print "Finished running doctest(musicalpitch)."

    a = MusicalPitch.new_from_notename("fis")
    nn = a.get_user_octave_notename()
    a = MusicalPitch.new_from_notename("g")
    b = MusicalPitch.new_from_notename("f")
    r = a - b
    print "%s - %s =" % (a, b), r
    r = a - 2
    print "%s - 2 =" % a, r
    r = a + 3
    print "%s + 3 =" % a, r
    print "a < b", a < b
    print "a > b", a > b
    print a, a.semitone_pitch()
    print "new_from_int", MusicalPitch.new_from_int(55) == a
    print "new_from_notename", MusicalPitch.new_from_notename(a.get_notename()) == a
    print "clone", a.clone() == a
